#
# Copyright (c) 2019 Nightwatch Cybersecurity.
#
# This file is part of truegaze
# (see https://github.com/nightwatchcybersecurity/truegaze).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import re

from androguard.core.bytecodes.apk import APK
from asn1crypto import cms
from cryptography.hazmat.primitives.asymmetric import utils
import click
from roca.detect import RocaFingerprinter as Roca

from truegaze.plugins.base import BasePlugin

# Regex pattern for the configuration file
CERTIFICATE_FILE_PATTERN =\
    re.compile(r'META-INF/.*\.RSA|META-INF/.*\.DSA')


# Plugin to check for weak Android signing keys and signatures
class WeakKeyPlugin(BasePlugin):
    name = 'WeakKeyPlugin'
    desc = 'Detection of weak Android signing keys'
    supports_android = True
    supports_ios = False

    # Main scanning method
    def scan(self):
        # Open the file
        apk = APK(self.filename)

        # Get certificates
        unique_certs = WeakKeyPlugin.get_certificates(apk)
        if len(unique_certs) == 0:
            click.echo('-- Cannot find the any certificates in the APK File, skipping')
            return

        # Loop through the certificates and check for weak keys
        messages = list()
        messages.extend(WeakKeyPlugin.check_for_short_keys(unique_certs))
        messages.extend(WeakKeyPlugin.check_for_roca(unique_certs))

        # Check for weak DSA signatures
        signatures = WeakKeyPlugin.get_signatures(apk)
        if len(signatures) > 1:
            messages = WeakKeyPlugin.check_for_weak_signatures(signatures)

        # Show results if needed
        if len(messages) > 0:
            click.echo("-- Found " + str(len(messages)) + ' issues')
            for message in messages:
                click.echo(message)
        else:
            click.echo("-- No issues found")

    # Get a list of certificates from the apk
    # TODO: add tests
    @staticmethod
    def get_certificates(apk):
        all_certs = list()
        all_certs.extend(apk.get_certificates_v1())
        all_certs.extend(apk.get_certificates_v2())
        all_certs.extend(apk.get_certificates_v3())

        unique_certs = dict()
        for cert in all_certs:
            unique_certs[cert.sha256_fingerprint] = cert

        return unique_certs.values()

    # Get a list of DSA/ECDSA signatures from the APK file
    # TODO: add tests
    @staticmethod
    def get_signatures(apk):
        signatures = list()
        certs_v1 = apk.get_certificates_v1()
        for x in range(len(certs_v1)):
            if certs_v1[x].public_key.algorithm == 'dsa' or certs_v1[x].public_key.algorithm == 'ecdsa':
                data = cms.ContentInfo.load(apk.get_signatures()[x])
                signatures.append(data['content']['signer_infos'][0]['signature'].contents)

        if apk._is_signed_v2:
            certs_v2 = apk.get_certificates_v2()
            for x in range(len(certs_v2)):
                if certs_v2[x].public_key.algorithm == 'dsa' or certs_v1[x].public_key.algorithm == 'ecdsa':
                    signatures.append(apk._v2_signing_data[x].signatures[0][1])

        if apk._is_signed_v3:
            certs_v3 = apk.get_certificates_v3()
            for x in range(len(certs_v3)):
                if certs_v3[x].public_key.algorithm == 'dsa' or certs_v1[x].public_key.algorithm == 'ecdsa':
                    signatures.append(apk._v3_signing_data[x].signatures[0][1])

        return signatures

    # Check for short keys
    # TODO: add tests
    @staticmethod
    def check_for_short_keys(certs):
        messages = []
        for cert in certs:
            if (cert.public_key.algorithm == 'rsa' and cert.public_key.bit_size < 2048) or \
                    (cert.public_key.algorithm == 'dsa' and cert.public_key.bit_size < 2048) or \
                    (cert.public_key.algorithm == 'ecdsa' and cert.public_key.bit_size < 224):
                messages.append('---- ISSUE (' +
                                'algorithm: ' + cert.public_key.algorithm +
                                ', fingerprint: ' + cert.sha1_fingerprint.replace(' ', '') +
                                '): Key is less than 2048 bits, size is ' + str(cert.public_key.bit_size) + ' bits')
        return messages

    # Check for ROCA attacks
    # TODO: add tests
    @staticmethod
    def check_for_roca(certs):
        messages = []
        for cert in certs:
            if cert.public_key.algorithm == 'rsa':
                modulus = cert.public_key.native['public_key']['modulus']
                if Roca().has_fingerprint_moduli(modulus) or Roca().has_fingerprint_dlog(modulus):
                    messages.append('---- ISSUE (' +
                                    'algorithm: ' + cert.public_key.algorithm +
                                    ', fingerprint: ' + cert.sha1_fingerprint.replace(' ', '') +
                                    '): Key vulnerable to ROCA attack - see https://roca.crocs.fi.muni.cz/')

        return messages

    # Check for weak DSA/ECDSA signatures
    # TODO: add tests
    @staticmethod
    def check_for_weak_signatures(signatures):
        messages = []

        # Extra r values from signatures
        values = list()
        for signature in signatures:
            (r, s) = utils.decode_dss_signature(signature)
            values.append(r)

        # Check if any appear more than once
        for r in values:
            if values.count(r) > 1:
                messages.append('---- ISSUE: DSA "r" value occurs more than once, private key is recoverable; k = ' + r)

        return messages



#
# Copyright (c) 2019 Nightwatch Cybersecurity.
#
# This file is part of truegaze
# (see https://github.com/nightwatchcybersecurity/truegaze).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import json
import re

import click
from jsonschema.validators import validator_for

from truegaze.plugins.base import BasePlugin
from truegaze.utils import TruegazeUtils

# Regex pattern for the configuration file
CONFIG_FILE_PATTERN =\
    re.compile(r'assets/ADBMobileConfig\.json|ADBMobileConfig\.json|.*/ADBMobileConfig\.json')


#
# Plugin to support detection of incorrect SSL configuration in the Adobe Mobile SDK. This plugin will not
# search for config files that are not named in a standard fashion.
#
# Adobe documentation including field definitions can be found here:
# https://docs.adobe.com/content/help/en/mobile-services/android/configuration-android/json-config.html
# https://docs.adobe.com/content/help/en/mobile-services/ios/config-ios/json-config.html
#


class AdobeMobileSdkPlugin(BasePlugin):
    name = 'AdobeMobileSdk'
    desc = 'Detection of incorrect SSL configuration\nin the Adobe Mobile SDK'
    supports_android = True
    supports_ios = True
    zip_file = None

    # Main scanning method
    def scan(self):
        # On Android, the config file is usually in the assets folder but can be placed elsewhere.
        # On iOS the configuration file can be anywhere.

        # Load file
        self.zip_file = TruegazeUtils.open_file_as_zip(self.filename)

        # Search all paths for the config file
        paths = AdobeMobileSdkPlugin.get_paths(self.zip_file)
        if len(paths) == 0:
            click.echo('-- No Adobe integration in this application (no "ADBMobileConfig.json" file); skipping test')
            return

        # Loop through files, parse the JSON and analyze
        click.echo('-- Found ' + str(len(paths)) + ' configuration file(s)')
        for path in paths:
            click.echo('-- Scanning "' + path + "'")

            # Try to parse the data
            parsed_data = AdobeMobileSdkPlugin.parse_data(self.zip_file, path)
            if not parsed_data:
                click.echo('---- ERROR: Unable to parse config file - will skip. File: ' + path)
                continue

            # Validate the file
            messages = AdobeMobileSdkPlugin.validate(parsed_data)
            if len(messages) > 0:
                click.echo("-- Found " + str(len(messages)) + ' issues')
                for message in messages:
                    click.echo(message)
            else:
                click.echo("-- No issues found")

    # Gets paths for the configuration file from the ZIP File
    @staticmethod
    def get_paths(zip_file):
        return TruegazeUtils.get_matching_paths_from_zip(zip_file, CONFIG_FILE_PATTERN)

    # Parses the config file from a given path
    @staticmethod
    def parse_data(zip_file, path):
        data = zip_file.read(path)
        try:
            parsed_data = json.loads(data.decode())
        except json.JSONDecodeError:
            return None
        return parsed_data

    # Validates the config file against the JSON schema
    @staticmethod
    def validate(parsed_data):
        # Load the schema
        schema_file = open('rules/adobe_mobile_sdk.schema', 'r')
        schema_data = json.load(schema_file)

        # Validate the file
        validator = validator_for(schema_data)
        errors = validator(schema=schema_data).iter_errors(parsed_data)

        # Extract error messages and return
        messages = []
        for error in errors:
            messages.append('---- ISSUE: ' + error.schema['title'] + '; ' + error.message)
        return messages



#
# Copyright (c) 2019 Nightwatch Cybersecurity.
#
# This file is part of truegaze
# (see https://github.com/nightwatchcybersecurity/truegaze).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import plistlib
import re
import zipfile

# Name of the Android manifest file
ANDROID_MANIFEST = 'AndroidManifest.xml'

# Regex pattern to use when searching iOS files
IOS_PATTERN = re.compile(r'Payload/[^/]*.\.app/Info.plist')


class TruegazeUtils(object):
    """Various utility functions, split off from the main class for ease of unit testing"""

    @staticmethod
    def get_version():
        """Gets the current version"""
        return "0.1.4"

    @staticmethod
    def open_file_as_zip(filename):
        """
        Tries to open the provided file as a zipfile

        :param filename: file to open
        :return: zipfile.ZipFile
        """
        try:
            return zipfile.ZipFile(filename, 'r')
        except (zipfile.BadZipfile, FileNotFoundError, zipfile.LargeZipFile):
            return None

    @staticmethod
    def get_android_manifest(zip_file):
        """
        Check if this is an Android application by looking for the AndroidManifest.xml file in root

        :param zip_file: zipfile.ZipFile to scan
        :return: path to the Android manifest file
        """
        try:
            file_info = zip_file.getinfo(ANDROID_MANIFEST)
            if file_info.file_size > 0:
                return ANDROID_MANIFEST
        except KeyError:
            return None

    @staticmethod
    def get_ios_manifest(zip_file):
        """
        Check if this is an iOS application by looking for the application and its plist

        :param zip_file: zipfile.ZipFile to scan
        :return: path to the iOS plist file
        """
        # IPA files have a /Payload/[something].app directory with the plist file in it, try to find it via regex
        paths = TruegazeUtils.get_matching_paths_from_zip(zip_file, IOS_PATTERN, True)

        # Check if the path was found and try to parse
        if len(paths) > 0:
            plist_path = paths[0]
            plist_contents = zip_file.read(plist_path)
            try:
                plist_dic = plistlib.loads(plist_contents)
            except plistlib.InvalidFileException:
                return None

            # Test to make sure some required keys are present
            if ('CFBundleDisplayName' in plist_dic) and \
                    ('CFBundleIdentifier' in plist_dic) and \
                    ('CFBundleShortVersionString' in plist_dic):
                return plist_path

        # Otherwise, return None if not detected
        return None

    @staticmethod
    def get_matching_paths_from_zip(zip_file, pattern, stop_after_first=False):
        """
        Searches ZIP file for list of matching paths

        :param zip_file: zipfile.ZipFile to scan
        :param pattern: regex pattern to use
        :param stop_after_first: whether to stop once first match is found
        :return: list of matched paths
        """
        file_list = zip_file.namelist()
        paths = []
        for file_path in file_list:
            matched = pattern.match(file_path)
            if matched is not None:
                paths.append(matched.group())
                if stop_after_first:
                    break

        return paths



from setuptools import find_packages, setup
from truegaze.utils import TruegazeUtils

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='truegaze',
    version=TruegazeUtils.get_version(),
    description='Static analysis tool for Android/iOS apps focusing on security issues outside the source code.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/nightwatchcybersecurity/truegaze',
    author='Nightwatch Cybersecurity',
    author_email='research@nightwatchcybersecurity.com',
    license='GNU',
    packages=find_packages(exclude=["scripts.*", "scripts", "tests.*", "tests"]),
    include_package_data=True,
    install_requires=open('requirements.txt').read().splitlines(),
    entry_points={
        'console_scripts': [
            'truegaze = truegaze.cli:cli'
        ]
    },
    classifiers=[
        'Environment :: Console',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>=3.5',
    project_urls={
        'Bug Reports': 'https://github.com/nightwatchcybersecurity/truegaze/issues',
        'Source': 'https://github.com/nightwatchcybersecurity/truegaze',
    },
)
